@@ -116,6 +116,9 @@
 /*	Request DNSSEC validation. This flag is silently ignored
 /*	when the system stub resolver API, resolver(3), does not
 /*	implement DNSSEC.
+/*	Automatically turns on the RES_TRUSTAD flag on systems that
+/*	support this flag (this behavior will be more configurable
+/*	in a later release).
 /* .RE
 /* .IP lflags
 /*	Flags that control the operation of the dns_lookup*()
@@ -145,6 +148,8 @@
 /*	available. The per-record reply TTL specifies how long the
 /*	DNS_NOTFOUND answer is valid. The caller should pass the
 /*	record(s) to dns_rr_free().
+/*	Logs a warning if the RES_DNSRCH or RES_DEFNAMES resolver
+/*	flags are set, and disables those flags.
 /* .RE
 /* .IP ltype
 /*	The resource record types to be looked up. In the case of
@@ -277,6 +282,43 @@ typedef struct DNS_REPLY {
 #define INET_ADDR_LEN	4		/* XXX */
 #define INET6_ADDR_LEN	16		/* XXX */
 
+ /*
+  * Use the theadsafe resolver API if available, not because it is theadsafe,
+  * but because it has more functionality.
+  */
+#ifdef USE_NRES_INIT
+static struct __res_state dns_res_state;
+
+#define DNS_RES_H_ERRNO(statp)	((statp)->res_h_errno)
+
+ /*
+  * Alias new resolver API calls to the legacy resolver API which stores
+  * resolver and error state in global variables.
+  */
+#else
+#define dns_res_state		_res
+#undef	res_ninit
+#undef	res_nmkquery
+#undef	res_nsearch
+#undef	res_nsend
+#define res_ninit(statp)	res_init()
+#define res_nmkquery(statp, op, dname, class, type, data, datalen, \
+		newrr, buf, buflen) \
+	res_mkquery((op), (dname), (class), (type), (data), (datalen), \
+		(newrr), (buf), (buflen))
+#define res_nsearch(statp, dname, class, type, answer, anslen) \
+	res_search((dname), (class), (type), (answer), (anslen))
+#define res_nsend(statp, msg, msglen, answer, anslen) \
+	res_send((msg), (msglen), (answer), (anslen))
+#define DNS_RES_H_ERRNO(statp)	(h_errno)
+#endif
+
+#ifdef USE_SET_H_ERRNO
+#define DNS_RES_SET_H_ERRNO(statp, err)	(set_h_errno(err))
+#else
+#define DNS_RES_SET_H_ERRNO(statp, err)	(DNS_RES_H_ERRNO(statp) = (err))
+#endif
+
  /*
   * To improve postscreen's whitelisting support, we need to know how long a
   * DNSBL "not found" answer is valid. The 2010 implementation assumed it was
@@ -338,34 +380,36 @@ static int dns_res_query(const char *name, int class, int type,
 #define NO_MKQUERY_DATA_LEN     ((int) 0)
 #define NO_MKQUERY_NEWRR        ((unsigned char *) 0)
 
-    if ((len = res_mkquery(QUERY, name, class, type, NO_MKQUERY_DATA_BUF,
-			   NO_MKQUERY_DATA_LEN, NO_MKQUERY_NEWRR,
-			   msg_buf, sizeof(msg_buf))) < 0) {
-	SET_H_ERRNO(NO_RECOVERY);
+    if ((len = res_nmkquery(&dns_res_state,
+			    QUERY, name, class, type, NO_MKQUERY_DATA_BUF,
+			    NO_MKQUERY_DATA_LEN, NO_MKQUERY_NEWRR,
+			    msg_buf, sizeof(msg_buf))) < 0) {
+	DNS_RES_SET_H_ERRNO(&dns_res_state, NO_RECOVERY);
 	if (msg_verbose)
-	    msg_info("res_mkquery() failed");
+	    msg_info("res_nmkquery() failed");
 	return (len);
-    } else if ((len = res_send(msg_buf, len, answer, anslen)) < 0) {
-	SET_H_ERRNO(TRY_AGAIN);
+    } else if ((len = res_nsend(&dns_res_state,
+				msg_buf, len, answer, anslen)) < 0) {
+	DNS_RES_SET_H_ERRNO(&dns_res_state, TRY_AGAIN);
 	if (msg_verbose)
-	    msg_info("res_send() failed");
+	    msg_info("res_nsend() failed");
 	return (len);
     } else {
 	switch (reply_header->rcode) {
 	case NXDOMAIN:
-	    SET_H_ERRNO(HOST_NOT_FOUND);
+	    DNS_RES_SET_H_ERRNO(&dns_res_state, HOST_NOT_FOUND);
 	    break;
 	case NOERROR:
 	    if (reply_header->ancount != 0)
-		SET_H_ERRNO(0);
+		DNS_RES_SET_H_ERRNO(&dns_res_state, 0);
 	    else
-		SET_H_ERRNO(NO_DATA);
+		DNS_RES_SET_H_ERRNO(&dns_res_state, NO_DATA);
 	    break;
 	case SERVFAIL:
-	    SET_H_ERRNO(TRY_AGAIN);
+	    DNS_RES_SET_H_ERRNO(&dns_res_state, TRY_AGAIN);
 	    break;
 	default:
-	    SET_H_ERRNO(NO_RECOVERY);
+	    DNS_RES_SET_H_ERRNO(&dns_res_state, NO_RECOVERY);
 	    break;
 	}
 	return (len);
@@ -399,18 +443,19 @@ static int dns_res_search(const char *name, int class, int type,
     if (keep_notfound)
 	/* Prepare for returning a null-padded server reply. */
 	memset(answer, 0, anslen);
-    len = res_search(name, class, type, answer, anslen);
+    len = res_nsearch(&dns_res_state, name, class, type, answer, anslen);
     /* Begin API creep workaround. */
-    if (len < 0 && h_errno == 0) {
-	SET_H_ERRNO(TRY_AGAIN);
-	msg_warn("res_query(\"%s\", %d, %d, %p, %d) returns %d with h_errno==0"
-		 " -- setting h_errno=TRY_AGAIN",
+    if (len < 0 && DNS_RES_H_ERRNO(&dns_res_state) == 0) {
+	DNS_RES_SET_H_ERRNO(&dns_res_state, TRY_AGAIN);
+	msg_warn("res_nsearch(state, \"%s\", %d, %d, %p, %d) returns %d"
+		 " with h_errno==0 -- setting h_errno=TRY_AGAIN",
 		 name, class, type, answer, anslen, len);
     }
     /* End API creep workaround. */
     if (len > 0) {
-	SET_H_ERRNO(0);
-    } else if (keep_notfound && NOT_FOUND_H_ERRNO(h_errno)) {
+	DNS_RES_SET_H_ERRNO(&dns_res_state, 0);
+    } else if (keep_notfound
+	       && NOT_FOUND_H_ERRNO(DNS_RES_H_ERRNO(&dns_res_state))) {
 	/* Expect to return a null-padded server reply. */
 	len = anslen;
     }
@@ -438,7 +483,8 @@ static int dns_query(const char *name, int type, unsigned flags,
     /*
      * Initialize the name service.
      */
-    if ((_res.options & RES_INIT) == 0 && res_init() < 0) {
+    if ((dns_res_state.options & RES_INIT) == 0
+	&& res_ninit(&dns_res_state) < 0) {
 	if (why)
 	    vstring_strcpy(why, "Name service initialization failure");
 	return (DNS_FAIL);
@@ -456,10 +502,20 @@ static int dns_query(const char *name, int type, unsigned flags,
     /*
      * Set extra options that aren't exposed to the application.
      */
-#define XTRA_FLAGS (RES_USE_EDNS0)
+#define XTRA_FLAGS (RES_USE_EDNS0 | RES_TRUSTAD)
 
     if (flags & RES_USE_DNSSEC)
-	flags |= RES_USE_EDNS0;
+	flags |= (RES_USE_EDNS0 | RES_TRUSTAD);
+
+    /*
+     * Can't append domains: we need the right SOA TTL.
+     */
+#define APPEND_DOMAIN_FLAGS (RES_DNSRCH | RES_DEFNAMES)
+
+    if (keep_notfound && (flags & APPEND_DOMAIN_FLAGS)) {
+	msg_warn("negative caching disables RES_DEFNAMES and RES_DNSRCH");
+	flags &= ~APPEND_DOMAIN_FLAGS;
+    }
 
     /*
      * Save and restore resolver options that we overwrite, to avoid
@@ -467,15 +523,15 @@ static int dns_query(const char *name, int type, unsigned flags,
      */
 #define SAVE_FLAGS (USER_FLAGS | XTRA_FLAGS)
 
-    saved_options = (_res.options & SAVE_FLAGS);
+    saved_options = (dns_res_state.options & SAVE_FLAGS);
 
     /*
      * Perform the lookup. Claim that the information cannot be found if and
      * only if the name server told us so.
      */
     for (;;) {
-	_res.options &= ~saved_options;
-	_res.options |= flags;
+	dns_res_state.options &= ~saved_options;
+	dns_res_state.options |= flags;
 	if (keep_notfound && var_dns_ncache_ttl_fix) {
 #ifdef HAVE_RES_SEND
 	    len = dns_res_query((char *) name, C_IN, type, reply->buf,
@@ -491,19 +547,21 @@ static int dns_query(const char *name, int type, unsigned flags,
 	    len = dns_res_search((char *) name, C_IN, type, reply->buf,
 				 reply->buf_len, keep_notfound);
 	}
-	_res.options &= ~flags;
-	_res.options |= saved_options;
+	dns_res_state.options &= ~flags;
+	dns_res_state.options |= saved_options;
 	reply_header = (HEADER *) reply->buf;
 	reply->rcode = reply_header->rcode;
-	if (h_errno != 0) {
+	if (DNS_RES_H_ERRNO(&dns_res_state) != 0) {
 	    if (why)
 		vstring_sprintf(why, "Host or domain name not found. "
 				"Name service error for name=%s type=%s: %s",
-			    name, dns_strtype(type), dns_strerror(h_errno));
+				name, dns_strtype(type), 
+				dns_strerror(DNS_RES_H_ERRNO(&dns_res_state)));
 	    if (msg_verbose)
 		msg_info("dns_query: %s (%s): %s",
-			 name, dns_strtype(type), dns_strerror(h_errno));
-	    switch (h_errno) {
+			 name, dns_strtype(type), 
+			dns_strerror(DNS_RES_H_ERRNO(&dns_res_state)));
+	    switch (DNS_RES_H_ERRNO(&dns_res_state)) {
 	    case NO_RECOVERY:
 		return (DNS_FAIL);
 	    case HOST_NOT_FOUND:
@@ -533,7 +591,7 @@ static int dns_query(const char *name, int type, unsigned flags,
      */
     if (len < 0)
 	msg_panic("dns_query: bad length %d (h_errno=%s)",
-		  len, dns_strerror(h_errno));
+		  len, dns_strerror(DNS_RES_H_ERRNO(&dns_res_state)));
 
     /*
      * Paranoia.
@@ -567,13 +625,13 @@ static int dns_query(const char *name, int type, unsigned flags,
      * Future proofing. If this reaches the panic call, then some code change
      * introduced a bug.
      */
-    if (h_errno == 0) {
+    if (DNS_RES_H_ERRNO(&dns_res_state) == 0) {
 	return (DNS_OK);
     } else if (keep_notfound) {
 	return (DNS_NOTFOUND);
     } else {
 	msg_panic("dns_query: unexpected reply status: %s",
-		  dns_strerror(h_errno));
+		  dns_strerror(DNS_RES_H_ERRNO(&dns_res_state)));
     }
 }
 
@@ -951,7 +1009,7 @@ int     dns_lookup_x(const char *name, unsigned type, unsigned flags,
 			    name);
 	if (rcode)
 	    *rcode = NXDOMAIN;
-	SET_H_ERRNO(HOST_NOT_FOUND);
+	DNS_RES_SET_H_ERRNO(&dns_res_state, HOST_NOT_FOUND);
 	return (DNS_NOTFOUND);
     }
 
@@ -965,7 +1023,7 @@ int     dns_lookup_x(const char *name, unsigned type, unsigned flags,
 			    name);
 	if (rcode)
 	    *rcode = NXDOMAIN;
-	SET_H_ERRNO(HOST_NOT_FOUND);
+	DNS_RES_SET_H_ERRNO(&dns_res_state, HOST_NOT_FOUND);
 	return (DNS_NOTFOUND);
     }
 
@@ -1015,7 +1073,7 @@ int     dns_lookup_x(const char *name, unsigned type, unsigned flags,
 	    if (why)
 		vstring_sprintf(why, "Domain %s does not accept mail (nullMX)",
 				name);
-	    SET_H_ERRNO(NO_DATA);
+	    DNS_RES_SET_H_ERRNO(&dns_res_state, NO_DATA);
 	    return (status);
 	case DNS_OK:
 	    if (rrlist && dns_rr_filter_maps) {
@@ -1084,7 +1142,7 @@ int     dns_lookup_rl(const char *name, unsigned flags, DNS_RR **rrlist,
 	    vstring_strcpy(hpref_rtext ? hpref_rtext : \
 			   (hpref_rtext = vstring_alloc(VSTRING_LEN(why))), \
 			   vstring_str(why)); \
-	hpref_h_errno = h_errno; \
+	hpref_h_errno = DNS_RES_H_ERRNO(&dns_res_state); \
     } while (0)
 
     /* Restore intermediate highest-priority result. */
@@ -1094,7 +1152,7 @@ int     dns_lookup_rl(const char *name, unsigned flags, DNS_RR **rrlist,
 	    *rcode = hpref_rcode; \
 	if (why && status != DNS_OK) \
 	    vstring_strcpy(why, vstring_str(hpref_rtext)); \
-	SET_H_ERRNO(hpref_h_errno); \
+	DNS_RES_SET_H_ERRNO(&dns_res_state, hpref_h_errno); \
     } while (0)
 
     if (rrlist)
